@@ -1,10 +1,7 @@
-import argparse, base58, enum, json, secp256k1, sys
+import argparse, base58, enum, json
 from functools import partial
 
-if 'src.ggmpc' in sys.modules:
-  ggmpc = sys.modules['src.ggmpc']
-elif 'ggmpc' in sys.modules:
-  ggmpc = sys.modules['ggmpc']
+from . import Ecdsa, Eddsa, curves, serialization
 
 class HelpFormatter(argparse.HelpFormatter):
   def add_argument(self, action):
@@ -26,17 +23,46 @@ def main():
 
   parser = argparse.ArgumentParser(prog='ggmpc', formatter_class=formatter)
 
-  subparsers = parser.add_subparsers(
+  command_subparsers = parser.add_subparsers(
     dest='COMMAND',
-    metavar='COMMAND',
     required=True,
   )
 
-  keyshare_parser = subparsers.add_parser(
+  ecdsa_parser = command_subparsers.add_parser(
+    'ecdsa',
+    help='ECDSA commands',
+  )
+  ecdsa_parser.add_argument(
+    '-c',
+    help='specify curve',
+    choices=['secp256k1'],
+    default='secp256k1',
+  )
+
+  eddsa_parser = command_subparsers.add_parser(
+    'eddsa',
+    help='EdDSA commands',
+  )
+  eddsa_parser.add_argument(
+    '-c',
+    help='specify curve',
+    choices=['ed25519'],
+    default='ed25519',
+  )
+
+  # ECDSA subcommands.
+
+  ecdsa_subparsers = ecdsa_parser.add_subparsers(
+    dest='SUBCOMMAND',
+    metavar='SUBCOMMAND',
+    required=True,
+  )
+
+  keyshare_parser = ecdsa_subparsers.add_parser(
     'keyshare',
     help='create key shares',
   )
-  keyshare_parser.set_defaults(func=command_key_share)
+  keyshare_parser.set_defaults(func=ecdsa_key_share)
   keyshare_parser.add_argument(
     '-i',
     type=int,
@@ -63,11 +89,11 @@ def main():
     default=True,
   )
 
-  keycombine_parser = subparsers.add_parser(
+  keycombine_parser = ecdsa_subparsers.add_parser(
     'keycombine',
     help='combine key shares',
   )
-  keycombine_parser.set_defaults(func=command_key_combine)
+  keycombine_parser.set_defaults(func=ecdsa_key_combine)
   keycombine_parser.add_argument(
     'KEYSHARE',
     type=str,
@@ -82,15 +108,15 @@ def main():
     default=True,
   )
 
-  signshare_parser = subparsers.add_parser(
+  signshare_parser = ecdsa_subparsers.add_parser(
     'signshare',
     help='create signing shares',
   )
-  signshare_parser.set_defaults(func=command_sign_share)
+  signshare_parser.set_defaults(func=ecdsa_sign_share)
   signshare_parser.add_argument(
     'KEYSHARE',
     type=str,
-    help='combined key shares for all signing players',
+    help='combined key shares for all signers',
     nargs='+',
   )
   signshare_parser.add_argument(
@@ -101,15 +127,15 @@ def main():
     default=True,
   )
 
-  signconvert_parser = subparsers.add_parser(
+  signconvert_parser = ecdsa_subparsers.add_parser(
     'signconvert',
     help='convert signing shares',
   )
-  signconvert_parser.set_defaults(func=command_sign_convert)
+  signconvert_parser.set_defaults(func=ecdsa_sign_convert)
   signconvert_parser.add_argument(
     'SIGNSHARE',
     type=str,
-    help='your signing share and shares received from other players',
+    help='your signing share and shares received from other signers',
     nargs='+',
   )
   signconvert_parser.add_argument(
@@ -120,15 +146,15 @@ def main():
     default=True,
   )
 
-  signcombine_parser = subparsers.add_parser(
+  signcombine_parser = ecdsa_subparsers.add_parser(
     'signcombine',
     help='combine converted signing shares and signature shares',
   )
-  signcombine_parser.set_defaults(func=command_sign_combine)
+  signcombine_parser.set_defaults(func=ecdsa_sign_combine)
   signcombine_parser.add_argument(
     'SIGNSHARE',
     type=str,
-    help='your signing shares and shares received from other players',
+    help='your signing shares and shares received from other signers',
     nargs='+',
   )
   signcombine_parser.add_argument(
@@ -139,11 +165,11 @@ def main():
     default=True,
   )
 
-  sign_parser = subparsers.add_parser(
+  sign_parser = ecdsa_subparsers.add_parser(
     'sign',
     help='sign message using converted signing shares',
   )
-  sign_parser.set_defaults(func=command_sign)
+  sign_parser.set_defaults(func=ecdsa_sign)
   sign_parser.add_argument(
     'MESSAGE',
     type=str,
@@ -153,7 +179,7 @@ def main():
   sign_parser.add_argument(
     'SIGNSHARE',
     type=str,
-    help='your combined signing share',
+    help='your combined signing shares',
     nargs='+',
   )
   sign_parser.add_argument(
@@ -164,11 +190,11 @@ def main():
     default=True,
   )
 
-  verify_parser = subparsers.add_parser(
+  verify_parser = ecdsa_subparsers.add_parser(
     'verify',
     help='verify a signature',
   )
-  verify_parser.set_defaults(func=command_verify)
+  verify_parser.set_defaults(func=ecdsa_verify)
   verify_parser.add_argument(
     'MESSAGE',
     type=str,
@@ -182,11 +208,157 @@ def main():
     nargs=1,
   )
 
-  deserialize_parser = subparsers.add_parser(
+  # EdDSA subcommands.
+
+  eddsa_subparsers = eddsa_parser.add_subparsers(
+    dest='SUBCOMMAND',
+    metavar='SUBCOMMAND',
+    required=True,
+  )
+
+  keyshare_parser = eddsa_subparsers.add_parser(
+    'keyshare',
+    help='create key shares',
+  )
+  keyshare_parser.set_defaults(func=eddsa_key_share)
+  keyshare_parser.add_argument(
+    '-i',
+    type=int,
+    help='player index',
+    required=True,
+  )
+  keyshare_parser.add_argument(
+    '-t',
+    type=int,
+    help='required threshold',
+    required=True,
+  )
+  keyshare_parser.add_argument(
+    '-n',
+    type=int,
+    help='number of players',
+    required=True,
+  )
+  keyshare_parser.add_argument(
+    '-nz',
+    help='disable compression',
+    dest='compress',
+    action='store_false',
+    default=True,
+  )
+
+  keycombine_parser = eddsa_subparsers.add_parser(
+    'keycombine',
+    help='combine key shares',
+  )
+  keycombine_parser.set_defaults(func=eddsa_key_combine)
+  keycombine_parser.add_argument(
+    'KEYSHARE',
+    type=str,
+    help='your key share and shares received from other players',
+    nargs='+',
+  )
+  keycombine_parser.add_argument(
+    '-nz',
+    help='disable compression',
+    dest='compress',
+    action='store_false',
+    default=True,
+  )
+
+  signshare_parser = eddsa_subparsers.add_parser(
+    'signshare',
+    help='create signing shares',
+  )
+  signshare_parser.set_defaults(func=eddsa_sign_share)
+  signshare_parser.add_argument(
+    'MESSAGE',
+    type=str,
+    help='message to sign',
+    nargs=1,
+  )
+  signshare_parser.add_argument(
+    'KEYSHARE',
+    type=str,
+    help='key shares for all signers',
+    nargs='+',
+  )
+  signshare_parser.add_argument(
+    '-nz',
+    help='disable compression',
+    dest='compress',
+    action='store_false',
+    default=True,
+  )
+
+  sign_parser = eddsa_subparsers.add_parser(
+    'sign',
+    help='sign message using signing shares',
+  )
+  sign_parser.set_defaults(func=eddsa_sign)
+  sign_parser.add_argument(
+    'MESSAGE',
+    type=str,
+    help='message to sign',
+    nargs=1,
+  )
+  sign_parser.add_argument(
+    'SIGNSHARE',
+    type=str,
+    help='signing shares',
+    nargs='+',
+  )
+  sign_parser.add_argument(
+    '-nz',
+    help='disable compression',
+    dest='compress',
+    action='store_false',
+    default=True,
+  )
+
+  signcombine_parser = eddsa_subparsers.add_parser(
+    'signcombine',
+    help='combine signature shares',
+  )
+  signcombine_parser.set_defaults(func=eddsa_sign_combine)
+  signcombine_parser.add_argument(
+    'SIGNSHARE',
+    type=str,
+    help='your signing shares and shares received from other signers',
+    nargs='+',
+  )
+  signcombine_parser.add_argument(
+    '-nz',
+    help='disable compression',
+    dest='compress',
+    action='store_false',
+    default=True,
+  )
+
+  verify_parser = eddsa_subparsers.add_parser(
+    'verify',
+    help='verify a signature',
+  )
+  verify_parser.set_defaults(func=eddsa_verify)
+  verify_parser.add_argument(
+    'MESSAGE',
+    type=str,
+    help='message to verify',
+    nargs=1,
+  )
+  verify_parser.add_argument(
+    'SIGNATURE',
+    type=str,
+    help='signature to verify',
+    nargs=1,
+  )
+
+  # Deserialize command
+
+  deserialize_parser = command_subparsers.add_parser(
     'deserialize',
     help='deserialize data',
   )
-  deserialize_parser.set_defaults(func=command_deserialize)
   deserialize_parser.add_argument(
     'DATA',
     type=str,
@@ -195,94 +367,162 @@ def main():
   )
 
   args = parser.parse_args()
-  args.func(args)
 
-def command_key_share(args):
+  if args.COMMAND == 'ecdsa':
+    args.func(Ecdsa(getattr(curves, args.c)), args)
+  elif args.COMMAND == 'eddsa':
+    args.func(Eddsa(getattr(curves, args.c)), args)
+  elif args.COMMAND == 'deserialize':
+    command_deserialize(args)
+  
+def ecdsa_key_share(mpc, args):
   ser = partial(serialize, compress=args.compress)
-  shares = ggmpc.key_share(args.i, args.t, args.n)
+  shares = mpc.key_share(args.i, args.t, args.n)
   P_i = shares[args.i]
-  print('\nSave for yourself: %s' % ser(DataType.P_SHARE, P_i))
+  print('\nSave for yourself: %s' % ser(DataType.ECDSA_P_SHARE, P_i))
   for i in shares:
     if i != args.i:
       print(
         '\nSend to player %d: %s' \
-        % (i, ser(DataType.N_SHARE, shares[i]))
+        % (i, ser(DataType.ECDSA_N_SHARE, shares[i]))
       )
 
-def command_key_combine(args):
+def ecdsa_key_combine(mpc, args):
   ser = partial(serialize, compress=args.compress)
-  shares = ggmpc.key_combine(list(map(deserialize, args.KEYSHARE)))
-  P_i = shares[next(filter(lambda i: 'p' in shares[i], shares))]
-  print('\nSave for yourself: %s' % ser(DataType.X_SHARE, P_i))
+  shares = mpc.key_combine(list(map(deserialize, args.KEYSHARE)))
+  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
+  print('\nSave for yourself: %s' % ser(DataType.ECDSA_X_SHARE, P_i))
   for i in shares:
     if i != P_i['i']:
       print(
         '\nSave for player %d: %s' \
-        % (i, ser(DataType.Y_SHARE, shares[i]))
+        % (i, ser(DataType.ECDSA_Y_SHARE, shares[i]))
       )
 
-def command_sign_share(args):
+def ecdsa_sign_share(mpc, args):
   ser = partial(serialize, compress=args.compress)
-  shares = ggmpc.sign_share(list(map(deserialize, args.KEYSHARE)))
-  P_i = shares[next(filter(lambda i: 'p' in shares[i], shares))]
-  print('\nSave for yourself: %s' % ser(DataType.W_SHARE, P_i))
+  shares = mpc.sign_share(list(map(deserialize, args.KEYSHARE)))
+  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
+  print('\nSave for yourself: %s' % ser(DataType.ECDSA_W_SHARE, P_i))
   for i in shares:
     if i != P_i['i']:
       print(
         '\nSend to player %d: %s' \
-        % (i, ser(DataType.K_SHARE, shares[i]))
+        % (i, ser(DataType.ECDSA_K_SHARE, shares[i]))
       )
 
-def command_sign_convert(args):
+def ecdsa_sign_convert(mpc, args):
   ser = partial(serialize, compress=args.compress)
   shares = list(map(deserialize, args.SIGNSHARE))
   for i in range(0, len(shares)):
     share = shares[i]
     if type(list(share.keys())[0]) == int:
-      shares[i] = shares[i][next(filter(lambda i: 'p' in share[i], share))]
-  shares = ggmpc.sign_convert(shares)
+      shares[i] = shares[i][next(filter(lambda i: not 'j' in share[i], share))]
+  shares = mpc.sign_convert(shares)
   P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
   P_j = shares[next(filter(lambda i: 'j' in shares[i], shares))]
   data_type = None
   if 'alpha' in P_i:
-    data_type = DataType.G_SHARE
+    data_type = DataType.ECDSA_G_SHARE
   else:
-    data_type = DataType.B_SHARE
+    data_type = DataType.ECDSA_B_SHARE
   print('\nSave for yourself:', ser(data_type, shares),)
   data_type = None
   if 'k' in P_j or 'alpha' in P_j:
     if 'k' in P_j:
-      data_type = DataType.A_SHARE
+      data_type = DataType.ECDSA_A_SHARE
     else:
-      data_type = DataType.M_SHARE
+      data_type = DataType.ECDSA_M_SHARE
     print('\nSend to player %d: %s' % (P_j['i'], ser(data_type, P_j)))
 
-def command_sign_combine(args):
+def ecdsa_sign_combine(mpc, args):
   ser = partial(serialize, compress=args.compress)
   shares = list(map(deserialize, args.SIGNSHARE))
   if 'r' in shares[0] and 's' in shares[0]:
-    print('\n%s' % ser(DataType.SIGNATURE, ggmpc.sign_combine(shares)))
+    print('\n%s' % ser(DataType.ECDSA_SIGNATURE, mpc.sign_combine(shares)))
   else:
-    shares = ggmpc.sign_combine(shares)
+    shares = mpc.sign_combine(shares)
     P_i = shares[next(filter(lambda i: 'k' in shares[i], shares))]
-    print('\nSave for yourself: %s' % ser(DataType.O_SHARE, P_i))
+    print('\nSave for yourself: %s' % ser(DataType.ECDSA_O_SHARE, P_i))
     for i in shares:
       if i != P_i['i']:
         print(
           '\nSend to player %d: %s' \
-          % (i, ser(DataType.D_SHARE, shares[i]))
+          % (i, ser(DataType.ECDSA_D_SHARE, shares[i]))
         )
-
-def command_sign(args):
+                                            
+def ecdsa_sign(mpc, args):
   ser = partial(serialize, compress=args.compress)
-  m = ggmpc.hash(args.MESSAGE[0].encode('ascii'))
+  M = args.MESSAGE[0].encode('ascii')
   shares = list(map(deserialize, args.SIGNSHARE))
-  print('\n%s' % ser(DataType.S_SHARE, ggmpc.sign(m, shares)))
+  print(
+    '\nSend to all players: %s' \
+    % ser(DataType.ECDSA_S_SHARE, mpc.sign(M, shares))
+  )
 
-def command_verify(args):
-  m = ggmpc.hash(args.MESSAGE[0].encode('ascii'))
-  assert ggmpc.verify(m, deserialize(args.SIGNATURE[0]))
-  print('\nverification succeeded')
+def ecdsa_verify(mpc, args):
+  M = args.MESSAGE[0].encode('ascii')
+  assert mpc.verify(M, deserialize(args.SIGNATURE[0]))
+  print('\nVerification succeeded')
+
+def eddsa_key_share(mpc, args):
+  ser = partial(serialize, compress=args.compress)
+  shares = mpc.key_share(args.i, args.t, args.n)
+  P_i = shares[args.i]
+  print('\nSave for yourself: %s' % ser(DataType.EDDSA_U_SHARE, P_i))
+  for i in shares:
+    if i != args.i:
+      print(
+        '\nSend to player %d: %s' \
+        % (i, ser(DataType.EDDSA_Y_SHARE, shares[i]))
+      )
+
+def eddsa_key_combine(mpc, args):
+  ser = partial(serialize, compress=args.compress)
+  shares = mpc.key_combine(list(map(deserialize, args.KEYSHARE)))
+  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
+  print('\nSave for yourself: %s' % ser(DataType.EDDSA_P_SHARE, P_i))
+  for i in shares:
+    if i != P_i['i']:
+      print(
+        '\nSave for player %d: %s' \
+        % (i, ser(DataType.EDDSA_J_SHARE, shares[i]))
+      )
+
+def eddsa_sign_share(mpc, args):
+  ser = partial(serialize, compress=args.compress)
+  M = args.MESSAGE[0].encode('ascii')
+  shares = mpc.sign_share(M, list(map(deserialize, args.KEYSHARE)))
+  P_i = shares[next(filter(lambda i: not 'j' in shares[i], shares))]
+  print('\nSave for yourself: %s' % ser(DataType.EDDSA_X_SHARE, P_i))
+  for i in shares:
+    if i != P_i['i']:
+      print(
+        '\nSend to player %d: %s' \
+        % (i, ser(DataType.EDDSA_R_SHARE, shares[i]))
+      )
+
+def eddsa_sign(mpc, args):
+  ser = partial(serialize, compress=args.compress)
+  M = args.MESSAGE[0].encode('ascii')
+  shares = list(map(deserialize, args.SIGNSHARE))
+  print(
+    '\nSend to all players: %s' \
+    % ser(DataType.EDDSA_G_SHARE, mpc.sign(M, shares))
+  )
+
+def eddsa_sign_combine(mpc, args):
+  ser = partial(serialize, compress=args.compress)
+  shares = list(map(deserialize, args.SIGNSHARE))
+  print(
+    '\n%s' \
+    % ser(DataType.EDDSA_SIGNATURE, mpc.sign_combine(shares))
+  )
+
+def eddsa_verify(mpc, args):
+  M = args.MESSAGE[0].encode('ascii')
+  assert mpc.verify(M, deserialize(args.SIGNATURE[0]))
+  print('\nVerification succeeded')
 
 def command_deserialize(args):
   print('\n%s' % json.dumps(encode(deserialize(args.DATA[0])), indent=2))
@@ -298,10 +538,13 @@ def encode(obj):
     elif key == 'j':
       encoded[key] = val
     elif key == 'y' or key == 'Gamma':
-      encoded[key] = ggmpc.serialize_point(val).hex()
+      if type(val) == int:
+        encoded[key] = serialization.serialize_int(val, 32).hex()
+      else:
+        encoded[key] = serialization.ecdsa.serialize_point(val).hex()
     else:
       length = (val.bit_length() + 7) // 8
-      encoded[key] = ggmpc.serialize_int(val, length).hex()
+      encoded[key] = serialization.serialize_int(val, length).hex()
   return encoded
 
 def serialize_to_json(obj):
@@ -320,49 +563,65 @@ def decode(encoded):
     elif key == 'j':
       obj[key] = val
     elif key == 'y' or key == 'Gamma':
-      obj[key] = ggmpc.point_deserialize(bytes.fromhex(val))[1]
+      obj[key] = serialization.point_deserialize(bytes.fromhex(val))[1]
     else:
-      obj[key] = ggmpc.deserialize_int(bytes.fromhex(val))[1]
+      obj[key] = serialization.deserialize_int(bytes.fromhex(val))[1]
   return obj
 
 def deserialize_from_json(string):
   return decode(json.loads(string))
 
 class DataType(enum.Enum):
-  P_SHARE = 'pshare', 'i0x'
-  N_SHARE = 'nshare', 'i0j1x'
-  X_SHARE = 'xshare', 'i0x'
-  Y_SHARE = 'yshare', 'i0j1x'
-  W_SHARE = 'wshare', 'i0x'
-  K_SHARE = 'kshare', 'i0j1x'
-  B_SHARE = 'bshare', 'i0j1x'
-  A_SHARE = 'ashare', 'i0j1x'
-  M_SHARE = 'mshare', 'i0j1x'
-  G_SHARE = 'gshare', 'i0j1x'
-  O_SHARE = 'oshare', 'i0x'
-  D_SHARE = 'dshare', 'i0j1x'
-  S_SHARE = 'sshare', 'i0x'
-  SIGNATURE = 'sigx',
+  ECDSA_P_SHARE = 'pshc', '_i0_'
+  ECDSA_N_SHARE = 'nshc', '_i0j1_'
+  ECDSA_X_SHARE = 'xshc', '_i0_'
+  ECDSA_Y_SHARE = 'yshc', '_i0j1_'
+  ECDSA_W_SHARE = 'wshc', '_i0_'
+  ECDSA_K_SHARE = 'kshc', '_i0j1_'
+  ECDSA_B_SHARE = 'bshc', '_i0j1_'
+  ECDSA_A_SHARE = 'ashc', '_i0j1_'
+  ECDSA_M_SHARE = 'mshc', '_i0j1_'
+  ECDSA_G_SHARE = 'gshc', '_i0j1_'
+  ECDSA_O_SHARE = 'oshc', '_i0_'
+  ECDSA_D_SHARE = 'dshc', '_i0j1_'
+  ECDSA_S_SHARE = 'sshc', '_i0_'
+  ECDSA_SIGNATURE = 'sigc_',
+  EDDSA_U_SHARE = 'ushd', '_i0_'
+  EDDSA_Y_SHARE = 'yshd', '_i0j1_'
+  EDDSA_P_SHARE = 'pshd', '_i0_'
+  EDDSA_J_SHARE = 'jshd', '_i0j1'
+  EDDSA_X_SHARE = 'xshd', '_i0_'
+  EDDSA_R_SHARE = 'rshd', '_i0j1_'
+  EDDSA_G_SHARE = 'gshd', '_i0_'
+  EDDSA_SIGNATURE = 'sigd_',
 
 alphabet = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
 
 def serialize(data_type, data, compress=True):
   if compress:
     serialize = {
-      DataType.P_SHARE: ggmpc.serialize_p_share,
-      DataType.N_SHARE: ggmpc.serialize_n_share,
-      DataType.X_SHARE: ggmpc.serialize_x_share,
-      DataType.Y_SHARE: ggmpc.serialize_y_share,
-      DataType.W_SHARE: ggmpc.serialize_w_share,
-      DataType.K_SHARE: ggmpc.serialize_k_share,
-      DataType.B_SHARE: ggmpc.serialize_b_share,
-      DataType.A_SHARE: ggmpc.serialize_a_share,
-      DataType.M_SHARE: ggmpc.serialize_m_share,
-      DataType.G_SHARE: ggmpc.serialize_g_share,
-      DataType.O_SHARE: ggmpc.serialize_o_share,
-      DataType.D_SHARE: ggmpc.serialize_d_share,
-      DataType.S_SHARE: ggmpc.serialize_s_share,
-      DataType.SIGNATURE: ggmpc.serialize_signature,
+      DataType.ECDSA_P_SHARE: serialization.ecdsa.serialize_p_share,
+      DataType.ECDSA_N_SHARE: serialization.ecdsa.serialize_n_share,
+      DataType.ECDSA_X_SHARE: serialization.ecdsa.serialize_x_share,
+      DataType.ECDSA_Y_SHARE: serialization.ecdsa.serialize_y_share,
+      DataType.ECDSA_W_SHARE: serialization.ecdsa.serialize_w_share,
+      DataType.ECDSA_K_SHARE: serialization.ecdsa.serialize_k_share,
+      DataType.ECDSA_B_SHARE: serialization.ecdsa.serialize_b_share,
+      DataType.ECDSA_A_SHARE: serialization.ecdsa.serialize_a_share,
+      DataType.ECDSA_M_SHARE: serialization.ecdsa.serialize_m_share,
+      DataType.ECDSA_G_SHARE: serialization.ecdsa.serialize_g_share,
+      DataType.ECDSA_O_SHARE: serialization.ecdsa.serialize_o_share,
+      DataType.ECDSA_D_SHARE: serialization.ecdsa.serialize_d_share,
+      DataType.ECDSA_S_SHARE: serialization.ecdsa.serialize_s_share,
+      DataType.ECDSA_SIGNATURE: serialization.ecdsa.serialize_signature,
+      DataType.EDDSA_U_SHARE: serialization.eddsa.serialize_u_share,
+      DataType.EDDSA_Y_SHARE: serialization.eddsa.serialize_y_share,
+      DataType.EDDSA_P_SHARE: serialization.eddsa.serialize_p_share,
+      DataType.EDDSA_J_SHARE: serialization.eddsa.serialize_j_share,
+      DataType.EDDSA_X_SHARE: serialization.eddsa.serialize_x_share,
+      DataType.EDDSA_R_SHARE: serialization.eddsa.serialize_r_share,
+      DataType.EDDSA_G_SHARE: serialization.eddsa.serialize_g_share,
+      DataType.EDDSA_SIGNATURE: serialization.eddsa.serialize_signature,
     }
     data = serialize[data_type](data)
     i, j = None, None
@@ -382,52 +641,61 @@ def serialize(data_type, data, compress=True):
       if j != None:
         index = len(data_type.value[0]) + data_type.value[1].find('1')
         prefix = prefix[:index] + alphabet[j - 1] + prefix[index + 1:]
-    val = 0
-    place = 1
-    for c in prefix[::-1]:
-      val += place * alphabet.index(c)
-      place *= 58
-    data += b'\x00\x00\x00\x00'
-    val *= pow(58, len(base58.b58encode(data)))
-    val += int.from_bytes(data, 'big')
-    data = val.to_bytes((val.bit_length() + 7) // 8, 'big')
-    return base58.b58encode_check(data[:-4]).decode('ascii')
+    if len(data) > 0:
+      return prefix + base58.b58encode_check(data).decode('ascii')
+    return prefix
   return serialize_to_json(data)
 
 def deserialize(data):
   if data[0] == '{':
     return deserialize_from_json(data)
   deserialize = {
-    DataType.P_SHARE: ggmpc.deserialize_p_share,
-    DataType.N_SHARE: ggmpc.deserialize_n_share,
-    DataType.X_SHARE: ggmpc.deserialize_x_share,
-    DataType.Y_SHARE: ggmpc.deserialize_y_share,
-    DataType.W_SHARE: ggmpc.deserialize_w_share,
-    DataType.K_SHARE: ggmpc.deserialize_k_share,
-    DataType.B_SHARE: ggmpc.deserialize_b_share,
-    DataType.A_SHARE: ggmpc.deserialize_a_share,
-    DataType.M_SHARE: ggmpc.deserialize_m_share,
-    DataType.G_SHARE: ggmpc.deserialize_g_share,
-    DataType.O_SHARE: ggmpc.deserialize_o_share,
-    DataType.D_SHARE: ggmpc.deserialize_d_share,
-    DataType.S_SHARE: ggmpc.deserialize_s_share,
-    DataType.SIGNATURE: ggmpc.deserialize_signature,
+    DataType.ECDSA_P_SHARE: serialization.ecdsa.deserialize_p_share,
+    DataType.ECDSA_N_SHARE: serialization.ecdsa.deserialize_n_share,
+    DataType.ECDSA_X_SHARE: serialization.ecdsa.deserialize_x_share,
+    DataType.ECDSA_Y_SHARE: serialization.ecdsa.deserialize_y_share,
+    DataType.ECDSA_W_SHARE: serialization.ecdsa.deserialize_w_share,
+    DataType.ECDSA_K_SHARE: serialization.ecdsa.deserialize_k_share,
+    DataType.ECDSA_B_SHARE: serialization.ecdsa.deserialize_b_share,
+    DataType.ECDSA_A_SHARE: serialization.ecdsa.deserialize_a_share,
+    DataType.ECDSA_M_SHARE: serialization.ecdsa.deserialize_m_share,
+    DataType.ECDSA_G_SHARE: serialization.ecdsa.deserialize_g_share,
+    DataType.ECDSA_O_SHARE: serialization.ecdsa.deserialize_o_share,
+    DataType.ECDSA_D_SHARE: serialization.ecdsa.deserialize_d_share,
+    DataType.ECDSA_S_SHARE: serialization.ecdsa.deserialize_s_share,
+    DataType.ECDSA_SIGNATURE: serialization.ecdsa.deserialize_signature,
+    DataType.EDDSA_U_SHARE: serialization.eddsa.deserialize_u_share,
+    DataType.EDDSA_Y_SHARE: serialization.eddsa.deserialize_y_share,
+    DataType.EDDSA_P_SHARE: serialization.eddsa.deserialize_p_share,
+    DataType.EDDSA_J_SHARE: serialization.eddsa.deserialize_j_share,
+    DataType.EDDSA_X_SHARE: serialization.eddsa.deserialize_x_share,
+    DataType.EDDSA_R_SHARE: serialization.eddsa.deserialize_r_share,
+    DataType.EDDSA_G_SHARE: serialization.eddsa.deserialize_g_share,
+    DataType.EDDSA_SIGNATURE: serialization.eddsa.deserialize_signature,
   }
   data_size = {
-    DataType.P_SHARE: 450,
-    DataType.N_SHARE: 451,
-    DataType.X_SHARE: 450,
-    DataType.Y_SHARE: 386,
-    DataType.W_SHARE: 514,
-    DataType.K_SHARE: 1154,
-    DataType.B_SHARE: 3267,
-    DataType.A_SHARE: 2690,
-    DataType.M_SHARE: 1538,
-    DataType.G_SHARE: 259,
-    DataType.O_SHARE: 163,
-    DataType.D_SHARE: 67,
-    DataType.S_SHARE: 98,
-    DataType.SIGNATURE: 97,
+    DataType.ECDSA_P_SHARE: 450,
+    DataType.ECDSA_N_SHARE: 451,
+    DataType.ECDSA_X_SHARE: 450,
+    DataType.ECDSA_Y_SHARE: 386,
+    DataType.ECDSA_W_SHARE: 514,
+    DataType.ECDSA_K_SHARE: 1154,
+    DataType.ECDSA_B_SHARE: 3267,
+    DataType.ECDSA_A_SHARE: 2690,
+    DataType.ECDSA_M_SHARE: 1538,
+    DataType.ECDSA_G_SHARE: 259,
+    DataType.ECDSA_O_SHARE: 163,
+    DataType.ECDSA_D_SHARE: 67,
+    DataType.ECDSA_S_SHARE: 98,
+    DataType.ECDSA_SIGNATURE: 97,
+    DataType.EDDSA_U_SHARE: 97,
+    DataType.EDDSA_Y_SHARE: 66,
+    DataType.EDDSA_P_SHARE: 97,
+    DataType.EDDSA_J_SHARE: 2,
+    DataType.EDDSA_X_SHARE: 129,
+    DataType.EDDSA_R_SHARE: 66,
+    DataType.EDDSA_G_SHARE: 97,
+    DataType.EDDSA_SIGNATURE: 96,
   }
   for data_type in DataType:
     if data[:len(data_type.value[0])] == data_type.value[0]:
@@ -441,25 +709,18 @@ def deserialize(data):
         index = prefix[len(data_type.value[0]):]
         if data_type.value[1].find('0') >= 0:
           i = alphabet.index(index[data_type.value[1].find('0')]) + 1
-          index_len += 1
         if data_type.value[1].find('1') >= 0:
           j = alphabet.index(index[data_type.value[1].find('1')]) + 1
-          index_len += 1
-      val = 0
-      place = 1
-      for c in prefix[::-1]:
-        val += place * alphabet.index(c)
-        place *= 58
-      val *= pow(58, len(data) - prefix_len)
-      data = base58.b58decode_check(data) + b'\x00\x00\x00\x00'
-      data = int.from_bytes(data, 'big') - val
-      data = data.to_bytes(data_size[data_type] - index_len + 4, 'big')
+      if len(data) > prefix_len:
+        data = base58.b58decode_check(data[prefix_len:])
+      else:
+        data = b''
       if j != None:
         data = j.to_bytes(1, 'big') + data
       if i != None:
         data = i.to_bytes(1, 'big') + data
       rem, data = deserialize[data_type](data)
-      assert rem == b'\x00\x00\x00\x00'
+      assert len(rem) == 0
       return data
 
 if __name__ == '__main__':
